package org.eclipse.jdt.internal.compiler.ast;

package .var.folders._j.vfbv187d0bb8b8w5spzgg8180000gn.T;

import org.eclipse.jdt.internal.compiler.ASTVisitor;
import org.eclipse.jdt.internal.compiler.ast.ASTNode;
import org.eclipse.jdt.internal.compiler.ast.Expression;
import org.eclipse.jdt.internal.compiler.ast.NullAnnotationMatching;
import org.eclipse.jdt.internal.compiler.ast.OperatorExpression;
import org.eclipse.jdt.internal.compiler.ast.Reference;
import org.eclipse.jdt.internal.compiler.ast.TypeReference;
import org.eclipse.jdt.internal.compiler.codegen.CodeStream;
import org.eclipse.jdt.internal.compiler.flow.FlowContext;
import org.eclipse.jdt.internal.compiler.flow.FlowInfo;
import org.eclipse.jdt.internal.compiler.flow.UnconditionalFlowInfo;
import org.eclipse.jdt.internal.compiler.impl.Constant;
import org.eclipse.jdt.internal.compiler.lookup.BlockScope;
import org.eclipse.jdt.internal.compiler.lookup.FieldBinding;
import org.eclipse.jdt.internal.compiler.lookup.LocalVariableBinding;
import org.eclipse.jdt.internal.compiler.lookup.Scope;
import org.eclipse.jdt.internal.compiler.lookup.TypeBinding;

public class InstanceOfExpression extends OperatorExpression {
  public Expression expression;
  
  public TypeReference type;
  
  public InstanceOfExpression(Expression expression, TypeReference type) {
    this.expression = expression;
    this.type = type;
    type.bits |= 0x40000000;
    this.bits |= 0x7C0;
    this.sourceStart = expression.sourceStart;
    this.sourceEnd = type.sourceEnd;
  }
  
  public FlowInfo analyseCode(BlockScope currentScope, FlowContext flowContext, FlowInfo flowInfo) {
    UnconditionalFlowInfo unconditionalFlowInfo;
    LocalVariableBinding local = this.expression.localVariableBinding();
    if (local != null && (local.type.tagBits & 0x2L) == 0L) {
      unconditionalFlowInfo = this.expression.analyseCode(currentScope, flowContext, flowInfo)
        .unconditionalInits();
      FlowInfo initsWhenTrue = unconditionalFlowInfo.copy();
      initsWhenTrue.markAsComparedEqualToNonNull(local);
      flowContext.recordUsingNullReference((Scope)currentScope, local, 
          (ASTNode)this.expression, 1025, (FlowInfo)unconditionalFlowInfo);
      return FlowInfo.conditional(initsWhenTrue, unconditionalFlowInfo.copy());
    } 
    if (this.expression instanceof Reference && (currentScope.compilerOptions()).enableSyntacticNullAnalysisForFields) {
      FieldBinding field = ((Reference)this.expression).lastFieldBinding();
      if (field != null && (field.type.tagBits & 0x2L) == 0L)
        flowContext.recordNullCheckedFieldReference((Reference)this.expression, 1); 
    } 
    return (FlowInfo)this.expression.analyseCode(currentScope, flowContext, (FlowInfo)unconditionalFlowInfo)
      .unconditionalInits();
  }
  
  public void generateCode(BlockScope currentScope, CodeStream codeStream, boolean valueRequired) {
    int pc = codeStream.position;
    this.expression.generateCode(currentScope, codeStream, true);
    codeStream.instance_of(this.type, this.type.resolvedType);
    if (valueRequired) {
      codeStream.generateImplicitConversion(this.implicitConversion);
    } else {
      codeStream.pop();
    } 
    codeStream.recordPositionsFrom(pc, this.sourceStart);
  }
  
  public StringBuffer printExpressionNoParenthesis(int indent, StringBuffer output) {
    this.expression.printExpression(indent, output).append(" instanceof ");
    return this.type.print(0, output);
  }
  
  public TypeBinding resolveType(BlockScope scope) {
    this.constant = Constant.NotAConstant;
    TypeBinding expressionType = this.expression.resolveType(scope);
    TypeBinding checkedType = this.type.resolveType(scope, true);
    if (expressionType != null && checkedType != null && this.type.hasNullTypeAnnotation(TypeReference.AnnotationPosition.ANY))
      if (!expressionType.isCompatibleWith(checkedType) || NullAnnotationMatching.analyse(checkedType, expressionType, -1).isAnyMismatch())
        scope.problemReporter().nullAnnotationUnsupportedLocation(this.type);  
    if (expressionType == null || checkedType == null)
      return null; 
    if (!checkedType.isReifiable()) {
      scope.problemReporter().illegalInstanceOfGenericType(checkedType, (ASTNode)this);
    } else if (checkedType.isValidBinding()) {
      if ((expressionType != TypeBinding.NULL && expressionType.isBaseType()) || 
        !checkCastTypesCompatibility((Scope)scope, checkedType, expressionType, null))
        scope.problemReporter().notCompatibleTypesError(this, expressionType, checkedType); 
    } 
    return this.resolvedType = (TypeBinding)TypeBinding.BOOLEAN;
  }
  
  public void tagAsUnnecessaryCast(Scope scope, TypeBinding castType) {
    if (this.expression.resolvedType != TypeBinding.NULL)
      scope.problemReporter().unnecessaryInstanceof(this, castType); 
  }
  
  public void traverse(ASTVisitor visitor, BlockScope scope) {
    if (visitor.visit(this, scope)) {
      this.expression.traverse(visitor, scope);
      this.type.traverse(visitor, scope);
    } 
    visitor.endVisit(this, scope);
  }
}
